@@ -13,25 +13,21 @@
 // See the License for the specific language governing permissions and
 // limitations under the License.
 
-#include "hwy/contrib/sort/vqsort.h"
+#include "hwy/contrib/sort/vqsort.h"  // VQSort
 
 #undef HWY_TARGET_INCLUDE
 #define HWY_TARGET_INCLUDE "hwy/contrib/sort/vqsort_i64d.cc"
 #include "hwy/foreach_target.h"  // IWYU pragma: keep
 
 // After foreach_target
-#include "hwy/contrib/sort/traits-inl.h"
 #include "hwy/contrib/sort/vqsort-inl.h"
 
 HWY_BEFORE_NAMESPACE();
 namespace hwy {
 namespace HWY_NAMESPACE {
 
-void SortI64Desc(int64_t* HWY_RESTRICT keys, size_t num,
-                 int64_t* HWY_RESTRICT buf) {
-  SortTag<int64_t> d;
-  detail::SharedTraits<detail::TraitsLane<detail::OrderDescending<int64_t>>> st;
-  Sort(d, st, keys, num, buf);
+void SortI64Desc(int64_t* HWY_RESTRICT keys, size_t num) {
+  return VQSortStatic(keys, num, SortDescending());
 }
 
 // NOLINTNEXTLINE(google-readability-namespace-comments)
@@ -45,9 +41,8 @@ namespace {
 HWY_EXPORT(SortI64Desc);
 }  // namespace
 
-void Sorter::operator()(int64_t* HWY_RESTRICT keys, size_t n,
-                        SortDescending) const {
-  HWY_DYNAMIC_DISPATCH(SortI64Desc)(keys, n, Get<int64_t>());
+void VQSort(int64_t* HWY_RESTRICT keys, size_t n, SortDescending) {
+  HWY_DYNAMIC_DISPATCH(SortI64Desc)(keys, n);
 }
 
 }  // namespace hwy
