@@ -13,13 +13,11 @@
 // See the License for the specific language governing permissions and
 // limitations under the License.
 
-#include <stddef.h>
-#include <stdint.h>
+#include <stdio.h>
 
 #include <algorithm>  // std::fill
 #include <bitset>
-
-#include "hwy/base.h"
+#include <string>
 
 #undef HWY_TARGET_INCLUDE
 #define HWY_TARGET_INCLUDE "highway_test.cc"
@@ -42,6 +40,7 @@ HWY_NOINLINE void TestCappedLimit(T /* tag */) {
   const size_t N = Lanes(d);
   if (kLimit < N) {
     auto lanes = AllocateAligned<T>(N);
+    HWY_ASSERT(lanes);
     std::fill(lanes.get(), lanes.get() + N, T{0});
     Store(Set(d, T{1}), d, lanes.get());
     for (size_t i = kLimit; i < N; ++i) {
@@ -82,7 +81,7 @@ static size_t* MaxLanesForSize(size_t sizeof_t) {
 
 struct TestMaxLanes {
   template <class T, class D>
-  HWY_NOINLINE void operator()(T /*unused*/, D d) {
+  HWY_NOINLINE void operator()(T /*unused*/, D d) const {
     const size_t N = Lanes(d);
     const size_t kMax = MaxLanes(d);  // for RVV, includes LMUL
     HWY_ASSERT(N <= kMax);
@@ -93,6 +92,58 @@ struct TestMaxLanes {
   }
 };
 
+class TestFracNLanes {
+ private:
+  template <int kNewPow2, class D>
+  using DWithPow2 =
+      Simd<TFromD<D>, D::template NewN<kNewPow2, HWY_MAX_LANES_D(D)>(),
+           kNewPow2>;
+
+  template <typename T1, size_t N1, int kPow2, typename T2, size_t N2>
+  static HWY_INLINE void DoTestFracNLanes(Simd<T1, N1, 0> /*d1*/,
+                                          Simd<T2, N2, kPow2> d2) {
+    using D2 = Simd<T2, N2, kPow2>;
+    static_assert(IsSame<T1, T2>(), "T1 and T2 should be the same type");
+    static_assert(N2 > HWY_MAX_BYTES, "N2 > HWY_MAX_BYTES should be true");
+    static_assert(HWY_MAX_LANES_D(D2) == N1,
+                  "HWY_MAX_LANES_D(D2) should be equal to N1");
+    static_assert(N1 <= HWY_LANES(T2), "N1 <= HWY_LANES(T2) should be true");
+
+    TestMaxLanes()(T2(), d2);
+  }
+
+#if HWY_TARGET != HWY_SCALAR
+  template <class T, HWY_IF_LANES_LE(4, HWY_LANES(T))>
+  static HWY_INLINE void DoTest4LanesWithPow3(T /*unused*/) {
+    // If HWY_LANES(T) >= 4 is true, do DoTestFracNLanes for the
+    // MaxLanes(d) == 4, kPow2 == 3 case
+    const Simd<T, 4, 0> d;
+    DoTestFracNLanes(d, DWithPow2<3, decltype(d)>());
+  }
+  template <class T, HWY_IF_LANES_GT(4, HWY_LANES(T))>
+  static HWY_INLINE void DoTest4LanesWithPow3(T /*unused*/) {
+    // If HWY_LANES(T) < 4, do nothing
+  }
+#endif
+
+ public:
+  template <class T>
+  HWY_NOINLINE void operator()(T /*unused*/) const {
+    const Simd<T, 1, 0> d1;
+    DoTestFracNLanes(d1, DWithPow2<1, decltype(d1)>());
+    DoTestFracNLanes(d1, DWithPow2<2, decltype(d1)>());
+    DoTestFracNLanes(d1, DWithPow2<3, decltype(d1)>());
+
+#if HWY_TARGET != HWY_SCALAR
+    const Simd<T, 2, 0> d2;
+    DoTestFracNLanes(d2, DWithPow2<2, decltype(d2)>());
+    DoTestFracNLanes(d2, DWithPow2<3, decltype(d2)>());
+
+    DoTest4LanesWithPow3(T());
+#endif
+  }
+};
+
 HWY_NOINLINE void TestAllMaxLanes() {
   ForAllTypes(ForPartialVectors<TestMaxLanes>());
 
@@ -109,6 +160,8 @@ HWY_NOINLINE void TestAllMaxLanes() {
       }
     }
   }
+
+  ForAllTypes(TestFracNLanes());
 }
 
 struct TestSet {
@@ -118,6 +171,7 @@ struct TestSet {
     const Vec<D> v0 = Zero(d);
     const size_t N = Lanes(d);
     auto expected = AllocateAligned<T>(N);
+    HWY_ASSERT(expected);
     std::fill(expected.get(), expected.get() + N, T{0});
     HWY_ASSERT_VEC_EQ(d, expected.get(), v0);
 
@@ -139,6 +193,9 @@ struct TestSet {
     // here, even though we already suppress warnings in Undefined.
     HWY_DIAGNOSTICS(push)
     HWY_DIAGNOSTICS_OFF(disable : 4700, ignored "-Wuninitialized")
+#if HWY_COMPILER_GCC_ACTUAL
+    HWY_DIAGNOSTICS_OFF(disable : 4701, ignored "-Wmaybe-uninitialized")
+#endif
     const Vec<D> vu = Undefined(d);
     Store(vu, d, expected.get());
     HWY_DIAGNOSTICS(pop)
@@ -215,7 +272,7 @@ HWY_NOINLINE void TestAllSignBit() {
   ForFloatTypes(ForPartialVectors<TestSignBitFloat>());
 }
 
-// inline to work around incorrect SVE codegen (only first 128 bits used).
+// TODO(b/287462770): inline to work around incorrect SVE codegen
 template <class D, class V>
 HWY_INLINE void AssertNaN(D d, VecArg<V> v, const char* file, int line) {
   using T = TFromD<D>;
@@ -297,6 +354,7 @@ struct TestNaN {
 
     // Reduction
     HWY_ASSERT_NAN(d, SumOfLanes(d, nan));
+    HWY_ASSERT_NAN(d, Set(d, ReduceSum(d, nan)));
 // TODO(janwas): re-enable after QEMU/Spike are fixed
 #if HWY_TARGET != HWY_RVV
     HWY_ASSERT_NAN(d, MinOfLanes(d, nan));
@@ -311,7 +369,7 @@ struct TestNaN {
     HWY_ASSERT_NAN(d, Min(v1, nan));
     HWY_ASSERT_NAN(d, Max(v1, nan));
 #elif HWY_TARGET <= HWY_NEON_WITHOUT_AES && HWY_ARCH_ARM_V7
-    // ARMv7 NEON returns NaN if any input is NaN.
+    // Armv7 NEON returns NaN if any input is NaN.
     HWY_ASSERT_NAN(d, Min(v1, nan));
     HWY_ASSERT_NAN(d, Max(v1, nan));
     HWY_ASSERT_NAN(d, Min(nan, v1));
@@ -460,6 +518,56 @@ HWY_NOINLINE void TestAllDFromV() {
   ForAllTypes(ForPartialVectors<TestDFromV>());
 }
 
+struct TestBlocks {
+  template <class T, class D>
+  HWY_NOINLINE void operator()(T /*unused*/, D d) {
+    const size_t N = Lanes(d);
+    const size_t num_of_blocks = Blocks(d);
+    static constexpr size_t kNumOfLanesPer16ByteBlk = 16 / sizeof(T);
+    HWY_ASSERT(num_of_blocks >= 1);
+    HWY_ASSERT(num_of_blocks <= d.MaxBlocks());
+    HWY_ASSERT(
+        num_of_blocks ==
+        ((N < kNumOfLanesPer16ByteBlk) ? 1 : (N / kNumOfLanesPer16ByteBlk)));
+  }
+};
+
+HWY_NOINLINE void TestAllBlocks() {
+  ForAllTypes(ForPartialVectors<TestDFromV>());
+}
+
+struct TestBlockDFromD {
+  template <class T, class D>
+  HWY_NOINLINE void operator()(T /*unused*/, D d) {
+    const BlockDFromD<decltype(d)> d_block;
+    static_assert(d_block.MaxBytes() <= 16,
+                  "d_block.MaxBytes() <= 16 must be true");
+    static_assert(d_block.MaxBytes() <= d.MaxBytes(),
+                  "d_block.MaxBytes() <= d.MaxBytes() must be true");
+    static_assert(d.MaxBytes() > 16 || d_block.MaxBytes() == d.MaxBytes(),
+                  "d_block.MaxBytes() == d.MaxBytes() must be true if "
+                  "d.MaxBytes() is less than or equal to 16");
+    static_assert(d.MaxBytes() < 16 || d_block.MaxBytes() == 16,
+                  "d_block.MaxBytes() == 16 must be true if d.MaxBytes() is "
+                  "greater than or equal to 16");
+    static_assert(
+        IsSame<Vec<decltype(d_block)>, decltype(ExtractBlock<0>(Zero(d)))>(),
+        "Vec<decltype(d_block)> should be the same vector type as "
+        "decltype(ExtractBlock<0>(Zero(d)))");
+    const size_t d_bytes = Lanes(d) * sizeof(T);
+    const size_t d_block_bytes = Lanes(d_block) * sizeof(T);
+    HWY_ASSERT(d_block_bytes >= 1);
+    HWY_ASSERT(d_block_bytes <= d_bytes);
+    HWY_ASSERT(d_block_bytes <= 16);
+    HWY_ASSERT(d_bytes > 16 || d_block_bytes == d_bytes);
+    HWY_ASSERT(d_bytes < 16 || d_block_bytes == 16);
+  }
+};
+
+HWY_NOINLINE void TestAllBlockDFromD() {
+  ForAllTypes(ForPartialVectors<TestBlockDFromD>());
+}
+
 // NOLINTNEXTLINE(google-readability-namespace-comments)
 }  // namespace HWY_NAMESPACE
 }  // namespace hwy
@@ -482,6 +590,8 @@ HWY_EXPORT_AND_TEST_P(HighwayTest, TestAllIsFinite);
 HWY_EXPORT_AND_TEST_P(HighwayTest, TestAllCopyAndAssign);
 HWY_EXPORT_AND_TEST_P(HighwayTest, TestAllGetLane);
 HWY_EXPORT_AND_TEST_P(HighwayTest, TestAllDFromV);
+HWY_EXPORT_AND_TEST_P(HighwayTest, TestAllBlocks);
+HWY_EXPORT_AND_TEST_P(HighwayTest, TestAllBlockDFromD);
 }  // namespace hwy
 
 #endif
