@@ -13,8 +13,6 @@
 // See the License for the specific language governing permissions and
 // limitations under the License.
 
-#include <stddef.h>
-#include <stdint.h>
 #include <string.h>
 
 #include <cmath>  // std::isfinite
@@ -83,6 +81,7 @@ struct TestBitCast {
     // Must return the same bits
     auto from_lanes = AllocateAligned<T>(Lanes(d));
     auto to_lanes = AllocateAligned<ToT>(Lanes(dto));
+    HWY_ASSERT(from_lanes && to_lanes);
     Store(vf, d, from_lanes.get());
     Store(vt, dto, to_lanes.get());
     HWY_ASSERT(
@@ -179,6 +178,159 @@ HWY_NOINLINE void TestAllBitCast() {
 #endif
 }
 
+template <class TTo>
+struct TestResizeBitCastToOneLaneVect {
+  template <typename T, class D>
+  HWY_NOINLINE void operator()(T /*unused*/, D d) {
+    const size_t N = Lanes(d);
+    if (N == 0) {
+      return;
+    }
+
+    auto from_lanes = AllocateAligned<T>(N);
+    HWY_ASSERT(from_lanes);
+    auto v = Iota(d, 1);
+    Store(v, d, from_lanes.get());
+
+    const size_t num_of_bytes_to_copy = HWY_MIN(N * sizeof(T), sizeof(TTo));
+
+    int8_t active_bits_mask_i8_arr[sizeof(TTo)] = {};
+    for (size_t i = 0; i < num_of_bytes_to_copy; i++) {
+      active_bits_mask_i8_arr[i] = int8_t{-1};
+    }
+
+    TTo active_bits_int_mask;
+    CopyBytes<sizeof(TTo)>(active_bits_mask_i8_arr, &active_bits_int_mask);
+
+    const FixedTag<TTo, 1> d_to;
+    TTo expected_bits = 0;
+    memcpy(&expected_bits, from_lanes.get(), num_of_bytes_to_copy);
+
+    const auto expected = Set(d_to, expected_bits);
+    const auto v_active_bits_mask = Set(d_to, active_bits_int_mask);
+
+    const auto actual_1 = And(v_active_bits_mask, ResizeBitCast(d_to, v));
+    const auto actual_2 = ZeroExtendResizeBitCast(d_to, d, v);
+
+    HWY_ASSERT_VEC_EQ(d_to, expected, actual_1);
+    HWY_ASSERT_VEC_EQ(d_to, expected, actual_2);
+  }
+};
+
+HWY_NOINLINE void TestAllResizeBitCastToOneLaneVect() {
+  ForAllTypes(ForPartialVectors<TestResizeBitCastToOneLaneVect<uint32_t>>());
+#if HWY_HAVE_INTEGER64
+  ForAllTypes(ForPartialVectors<TestResizeBitCastToOneLaneVect<uint64_t>>());
+#endif
+}
+
+// Cast and ensure bytes are the same. Called directly from
+// TestAllSameSizeResizeBitCast or via TestSameSizeResizeBitCastFrom.
+template <typename ToT>
+struct TestSameSizeResizeBitCast {
+  template <typename T, class D>
+  HWY_NOINLINE void operator()(T /*unused*/, D d) {
+    const Repartition<ToT, D> dto;
+    const auto v = Iota(d, 1);
+    const auto expected = BitCast(dto, v);
+
+    const VFromD<decltype(dto)> actual_1 = ResizeBitCast(dto, v);
+    const VFromD<decltype(dto)> actual_2 = ZeroExtendResizeBitCast(dto, d, v);
+
+    HWY_ASSERT_VEC_EQ(dto, expected, actual_1);
+    HWY_ASSERT_VEC_EQ(dto, expected, actual_2);
+  }
+};
+
+// From D to all types.
+struct TestSameSizeResizeBitCastFrom {
+  template <typename T, class D>
+  HWY_NOINLINE void operator()(T t, D d) {
+    TestSameSizeResizeBitCast<uint8_t>()(t, d);
+    TestSameSizeResizeBitCast<uint16_t>()(t, d);
+    TestSameSizeResizeBitCast<uint32_t>()(t, d);
+#if HWY_HAVE_INTEGER64
+    TestSameSizeResizeBitCast<uint64_t>()(t, d);
+#endif
+    TestSameSizeResizeBitCast<int8_t>()(t, d);
+    TestSameSizeResizeBitCast<int16_t>()(t, d);
+    TestSameSizeResizeBitCast<int32_t>()(t, d);
+#if HWY_HAVE_INTEGER64
+    TestSameSizeResizeBitCast<int64_t>()(t, d);
+#endif
+    TestSameSizeResizeBitCast<float>()(t, d);
+#if HWY_HAVE_FLOAT64
+    TestSameSizeResizeBitCast<double>()(t, d);
+#endif
+  }
+};
+
+HWY_NOINLINE void TestAllSameSizeResizeBitCast() {
+  // For HWY_SCALAR and partial vectors, we can only cast to same-sized types:
+  // the former can't partition its single lane, and the latter can be smaller
+  // than a destination type.
+  const ForPartialVectors<TestSameSizeResizeBitCast<uint8_t>> to_u8;
+  to_u8(uint8_t());
+  to_u8(int8_t());
+
+  const ForPartialVectors<TestSameSizeResizeBitCast<int8_t>> to_i8;
+  to_i8(uint8_t());
+  to_i8(int8_t());
+
+  const ForPartialVectors<TestSameSizeResizeBitCast<uint16_t>> to_u16;
+  to_u16(uint16_t());
+  to_u16(int16_t());
+
+  const ForPartialVectors<TestSameSizeResizeBitCast<int16_t>> to_i16;
+  to_i16(uint16_t());
+  to_i16(int16_t());
+
+  const ForPartialVectors<TestSameSizeResizeBitCast<uint32_t>> to_u32;
+  to_u32(uint32_t());
+  to_u32(int32_t());
+  to_u32(float());
+
+  const ForPartialVectors<TestSameSizeResizeBitCast<int32_t>> to_i32;
+  to_i32(uint32_t());
+  to_i32(int32_t());
+  to_i32(float());
+
+#if HWY_HAVE_INTEGER64
+  const ForPartialVectors<TestSameSizeResizeBitCast<uint64_t>> to_u64;
+  to_u64(uint64_t());
+  to_u64(int64_t());
+#if HWY_HAVE_FLOAT64
+  to_u64(double());
+#endif
+
+  const ForPartialVectors<TestSameSizeResizeBitCast<int64_t>> to_i64;
+  to_i64(uint64_t());
+  to_i64(int64_t());
+#if HWY_HAVE_FLOAT64
+  to_i64(double());
+#endif
+#endif  // HWY_HAVE_INTEGER64
+
+  const ForPartialVectors<TestSameSizeResizeBitCast<float>> to_float;
+  to_float(uint32_t());
+  to_float(int32_t());
+  to_float(float());
+
+#if HWY_HAVE_FLOAT64
+  const ForPartialVectors<TestSameSizeResizeBitCast<double>> to_double;
+  to_double(double());
+#if HWY_HAVE_INTEGER64
+  to_double(uint64_t());
+  to_double(int64_t());
+#endif  // HWY_HAVE_INTEGER64
+#endif  // HWY_HAVE_FLOAT64
+
+#if HWY_TARGET != HWY_SCALAR
+  // For non-scalar vectors, we can cast all types to all.
+  ForAllTypes(ForGEVectors<64, TestSameSizeResizeBitCastFrom>());
+#endif
+}
+
 template <typename ToT>
 struct TestPromoteTo {
   template <typename T, class D>
@@ -234,6 +386,21 @@ HWY_NOINLINE void TestAllPromoteTo() {
 
   const ForPromoteVectors<TestPromoteTo<int64_t>, 1> to_i64div2;
   to_i64div2(int32_t());
+  to_i64div2(uint32_t());
+
+  const ForPromoteVectors<TestPromoteTo<uint64_t>, 2> to_u64div4;
+  to_u64div4(uint16_t());
+
+  const ForPromoteVectors<TestPromoteTo<int64_t>, 2> to_i64div4;
+  to_i64div4(int16_t());
+  to_i64div4(uint16_t());
+
+  const ForPromoteVectors<TestPromoteTo<uint64_t>, 3> to_u64div8;
+  to_u64div8(uint8_t());
+
+  const ForPromoteVectors<TestPromoteTo<int64_t>, 3> to_i64div8;
+  to_i64div8(int8_t());
+  to_i64div8(uint8_t());
 #endif
 
 #if HWY_HAVE_FLOAT64
@@ -272,7 +439,7 @@ AlignedFreeUniquePtr<float[]> F16TestCases(D d, size_t& padded) {
       2.00390625f, 3.99609375f,
       // negative +/- delta
       -2.00390625f, -3.99609375f,
-      // No infinity/NaN - implementation-defined due to ARM.
+      // No infinity/NaN - implementation-defined due to Arm.
   };
   constexpr size_t kNumTestCases = sizeof(test_cases) / sizeof(test_cases[0]);
   const size_t N = Lanes(d);
@@ -282,7 +449,8 @@ AlignedFreeUniquePtr<float[]> F16TestCases(D d, size_t& padded) {
   auto expected = AllocateAligned<float>(padded);
   size_t i = 0;
   for (; i < kNumTestCases; ++i) {
-    in[i] = test_cases[i];
+    // Ensure the value can be exactly represented as binary16.
+    in[i] = F32FromF16(F16FromF32(test_cases[i]));
   }
   for (; i < padded; ++i) {
     in[i] = 0.0f;
@@ -290,26 +458,38 @@ AlignedFreeUniquePtr<float[]> F16TestCases(D d, size_t& padded) {
   return in;
 }
 
+// This minimal interface is always supported, even if !HWY_HAVE_FLOAT16.
 struct TestF16 {
   template <typename TF32, class DF32>
-  HWY_NOINLINE void operator()(TF32 /*t*/, DF32 d32) {
-#if HWY_HAVE_FLOAT16
+  HWY_NOINLINE void operator()(TF32 /*t*/, DF32 df32) {
     size_t padded;
-    const size_t N = Lanes(d32);  // same count for f16
+    const size_t N = Lanes(df32);  // same count for f16
     HWY_ASSERT(N != 0);
-    auto in = F16TestCases(d32, padded);
-    using TF16 = float16_t;
-    const Rebind<TF16, DF32> d16;
+    auto in = F16TestCases(df32, padded);
+
+    using TF16 = hwy::float16_t;
+    const Rebind<TF16, DF32> df16;
+    const RebindToUnsigned<decltype(df16)> du16;
+    // Extra Load/Store to ensure they are usable.
     auto temp16 = AllocateAligned<TF16>(N);
 
-    for (size_t i = 0; i < padded; i += N) {
-      const auto loaded = Load(d32, &in[i]);
-      Store(DemoteTo(d16, loaded), d16, temp16.get());
-      HWY_ASSERT_VEC_EQ(d32, loaded, PromoteTo(d32, Load(d16, temp16.get())));
-    }
+    // Extra Zero/BitCast to ensure they are usable. Neg is tested in
+    // arithmetic_test.
+    const Vec<decltype(du16)> v0_u16 = BitCast(du16, Zero(df16));
+#if HWY_TARGET == HWY_SCALAR
+    const Vec<DF32> v0 = BitCast(df32, ZipLower(v0_u16, v0_u16));
 #else
-    (void)d32;
+    const Vec<DF32> v0 =
+        BitCast(df32, ZeroExtendVector(Twice<decltype(du16)>(), v0_u16));
 #endif
+
+    for (size_t i = 0; i < padded; i += N) {
+      const Vec<DF32> loaded = Or(Load(df32, &in[i]), v0);
+      const Vec<decltype(df16)> v16 = DemoteTo(df16, loaded);
+      Store(v16, df16, temp16.get());
+      HWY_ASSERT_VEC_EQ(df32, loaded,
+                        PromoteTo(df32, Load(df16, temp16.get())));
+    }
   }
 };
 
@@ -354,7 +534,6 @@ AlignedFreeUniquePtr<float[]> BF16TestCases(D d, size_t& padded) {
 struct TestBF16 {
   template <typename TF32, class DF32>
   HWY_NOINLINE void operator()(TF32 /*t*/, DF32 d32) {
-#if !defined(HWY_EMULATE_SVE)
     size_t padded;
     auto in = BF16TestCases(d32, padded);
     using TBF16 = bfloat16_t;
@@ -376,9 +555,6 @@ struct TestBF16 {
       const auto v16_loaded = Load(dbf16_half, temp16.get());
       HWY_ASSERT_VEC_EQ(d32, loaded, PromoteTo(d32, v16_loaded));
     }
-#else
-    (void)d32;
-#endif
   }
 };
 
@@ -435,12 +611,58 @@ HWY_NOINLINE void TestAllTruncate() {
   ForUnsignedTypes(ForPartialVectors<TestTruncateTo>());
 }
 
-// Separate function to attempt to work around a compiler bug on ARM: when this
+struct TestOrderedTruncate2To {
+  template <typename T, class D>
+  HWY_NOINLINE void operator()(T /*t*/, D d) {
+#if HWY_TARGET != HWY_SCALAR
+    const Repartition<MakeNarrow<T>, decltype(d)> dn;
+    using TN = TFromD<decltype(dn)>;
+
+    const size_t N = Lanes(d);
+    const size_t twiceN = N * 2;
+    auto from = AllocateAligned<T>(twiceN);
+    auto expected = AllocateAligned<TN>(twiceN);
+
+    const T max = LimitsMax<TN>();
+
+    constexpr uint32_t iota_base = 0xFA578D00;
+    const auto src_iota_a = Iota(d, static_cast<T>(iota_base));
+    const auto src_iota_b = Iota(d, static_cast<T>(iota_base + N));
+    const auto expected_iota_trunc_result =
+        Iota(dn, static_cast<TN>(iota_base));
+    const auto actual_iota_trunc_result =
+        OrderedTruncate2To(dn, src_iota_a, src_iota_b);
+    HWY_ASSERT_VEC_EQ(dn, expected_iota_trunc_result, actual_iota_trunc_result);
+
+    RandomState rng;
+    for (size_t rep = 0; rep < AdjustedReps(1000); ++rep) {
+      for (size_t i = 0; i < twiceN; ++i) {
+        const uint64_t bits = rng();
+        CopyBytes<sizeof(T)>(&bits, &from[i]);  // not same size
+        expected[i] = static_cast<TN>(from[i] & max);
+      }
+
+      const auto in_1 = Load(d, from.get());
+      const auto in_2 = Load(d, from.get() + N);
+      const auto actual = OrderedTruncate2To(dn, in_1, in_2);
+      HWY_ASSERT_VEC_EQ(dn, expected.get(), actual);
+    }
+#else
+    (void)d;
+#endif
+  }
+};
+
+HWY_NOINLINE void TestAllOrderedTruncate2To() {
+  ForU163264(ForShrinkableVectors<TestOrderedTruncate2To>());
+}
+
+// Separate function to attempt to work around a compiler bug on Arm: when this
 // is merged with TestIntFromFloat, outputs match a previous Iota(-(N+1)) input.
 struct TestIntFromFloatHuge {
   template <typename TF, class DF>
   HWY_NOINLINE void operator()(TF /*unused*/, const DF df) {
-    // The ARMv7 manual says that float->int saturates, i.e. chooses the
+    // The Armv7 manual says that float->int saturates, i.e. chooses the
     // nearest representable value. This works correctly on armhf with GCC, but
     // not with clang. For reasons unknown, MSVC also runs into an out-of-memory
     // error here.
@@ -555,8 +777,9 @@ class TestIntFromFloat {
 };
 
 HWY_NOINLINE void TestAllIntFromFloat() {
-  ForFloatTypes(ForPartialVectors<TestIntFromFloatHuge>());
-  ForFloatTypes(ForPartialVectors<TestIntFromFloat>());
+  // std::isfinite does not support float16_t.
+  ForFloat3264Types(ForPartialVectors<TestIntFromFloatHuge>());
+  ForFloat3264Types(ForPartialVectors<TestIntFromFloat>());
 }
 
 struct TestFloatFromInt {
@@ -594,8 +817,8 @@ struct TestFloatFromUint {
 
     // Integer positive
     HWY_ASSERT_VEC_EQ(df, Iota(df, TF(4.0)), ConvertTo(df, Iota(du, TU(4))));
-    HWY_ASSERT_VEC_EQ(df, Iota(df, TF(65535.0)),
-                      ConvertTo(df, Iota(du, 65535)));  // 2^16-1
+    HWY_ASSERT_VEC_EQ(df, Iota(df, TF(32767.0)),
+                      ConvertTo(df, Iota(du, 32767)));  // 2^16-1
     if (sizeof(TF) > 4) {
       HWY_ASSERT_VEC_EQ(df, Iota(df, TF(4294967295.0)),
                         ConvertTo(df, Iota(du, 4294967295ULL)));  // 2^32-1
@@ -655,7 +878,6 @@ HWY_NOINLINE void TestAllI32F64() {
 #endif
 }
 
-
 // NOLINTNEXTLINE(google-readability-namespace-comments)
 }  // namespace HWY_NAMESPACE
 }  // namespace hwy
@@ -667,11 +889,14 @@ namespace hwy {
 HWY_BEFORE_TEST(HwyConvertTest);
 HWY_EXPORT_AND_TEST_P(HwyConvertTest, TestAllRebind);
 HWY_EXPORT_AND_TEST_P(HwyConvertTest, TestAllBitCast);
+HWY_EXPORT_AND_TEST_P(HwyConvertTest, TestAllResizeBitCastToOneLaneVect);
+HWY_EXPORT_AND_TEST_P(HwyConvertTest, TestAllSameSizeResizeBitCast);
 HWY_EXPORT_AND_TEST_P(HwyConvertTest, TestAllPromoteTo);
 HWY_EXPORT_AND_TEST_P(HwyConvertTest, TestAllF16);
 HWY_EXPORT_AND_TEST_P(HwyConvertTest, TestAllBF16);
 HWY_EXPORT_AND_TEST_P(HwyConvertTest, TestAllConvertU8);
 HWY_EXPORT_AND_TEST_P(HwyConvertTest, TestAllTruncate);
+HWY_EXPORT_AND_TEST_P(HwyConvertTest, TestAllOrderedTruncate2To);
 HWY_EXPORT_AND_TEST_P(HwyConvertTest, TestAllIntFromFloat);
 HWY_EXPORT_AND_TEST_P(HwyConvertTest, TestAllFloatFromInt);
 HWY_EXPORT_AND_TEST_P(HwyConvertTest, TestAllFloatFromUint);
